import datetime
import os
import re
from collections import OrderedDict
from typing import List, Dict, Tuple, Optional

from metafix.constants import ReleaseCategory


def unique(seq: List):
    sorted_items = []
    out_items = []

    for item in seq:
        if not hasattr(item, "__iter__") or isinstance(item, str):
            if item not in out_items:
                out_items.append(item)
        else:
            s_item = sorted(item)
            if s_item not in sorted_items:
                sorted_items.append(s_item)
                out_items.append(item)
    return out_items


def normalize_track_title(str_in):
    str_in = ' '.join(str_in.replace("/", " / ").split())
    str_in = re.sub("(?i)( )?vs(.)?( )?", " vs. ", str_in)
    return normalize_str(str_in)

def normalize_release_title(str_in):
    return normalize_str(str_in)

def normalize_artist_name(str_in):
    str_in = re.sub("(?i)( )?vs(.)?( )?", " vs. ", str_in)
    return normalize_str(str_in)


def normalize_str(music_str):
    if not music_str:
        return ""
    music_str = re.sub("(?i) feat(.)?( )?", " feat. ", music_str)
    music_str = re.sub("(?i)\\(feat(.)?( )?", "(feat. ", music_str)
    music_str = re.sub(r'(?i) ?[\[{(]?(disc|disk|cd) ?(\d{1,2})[\]})]?', "", music_str)
    music_str = music_str.strip()

    return music_str


def flatten_artists(artists: List[str]) -> str:

    if len(artists) == 1:
        return artists[0]
    elif len(artists) == 2:
        return "{0} & {1}".format(artists[0], artists[1])
    else:
        return "{0} & {1}".format(", ".join(artists[0:-2]), artists[-1])


def split_release_title(release_title_full):
    release_title = release_title_full
    release_edition = ""

    edition_match = re.search(r"([({[][\w\-_ ]+[)}\]])$", release_title_full)
    if edition_match:
        release_edition = edition_match.groups()[0]
        release_title = release_title_full.replace(release_edition, "").strip()

    return release_title, release_edition


def get_category_fix_name(release: "Release") -> Tuple[str, Optional[ReleaseCategory]]:
    clean_name = re.sub(r"[\[{( ]web[\]})]?", "",
                        release.tracks[next(iter(release.tracks))].release_title, flags=re.I)

    lower_name = clean_name.lower()

    if lower_name.endswith(" ep"):
        return clean_name[0:-3].strip(), ReleaseCategory.EP
    elif lower_name.endswith("(ep)") or lower_name.endswith("[ep]") or lower_name.endswith("{ep}"):
        return clean_name[0:-4].strip(), ReleaseCategory.EP
    elif lower_name.endswith(" single"):
        return clean_name[0:-7].strip(), ReleaseCategory.SINGLE
    elif lower_name.endswith(" cds"):
        return clean_name[0:-4].strip(), ReleaseCategory.SINGLE
    elif lower_name.endswith(" (cds)") or lower_name.endswith("(cds)") or lower_name.endswith("{cds}"):
        return clean_name[0:-6].strip(), ReleaseCategory.SINGLE

    return clean_name, None


def tag_filter(tag: str, ignore_substrings: List[str], capitalize: bool) -> Optional[str]:
    """Return none for a bad tag, otherwise return a cleaned up version of the tag"""

    if len(tag) > 100:
        return None

    for curr_ignore in ignore_substrings:
        if tag.lower() in curr_ignore.lower() or curr_ignore.lower() in tag.lower():
            return None

    if tag.isdigit() and len(tag) == 4:
        return None

    if tag.lower() in ['seen live', 'favourites', 'all', 'awesome', 'love', 'spotify', 'favorite', 'favourite', 'fun',
                       'check out', 'sexy', 'amazing',
                       'genius', 'dj', 'want to see live', 'shit', 'officially shit',
                       'lesser known yet streamable artists', '<3', 'crap',
                       'bands i\'ve seen live', 'seen in concert', 'listen', 'good music', 'saw live', 'local',
                       'fuck off', 'hipster garbage', 'mp3',

                       'masterpiece', 'laptop', 'wishlist', 'cds i own', 'beautiful', 'epic', 'classic',
                       'records i own', 'playlist', 'cd', 'cool', 'great',
                       'good', 'own', 'emusic', 'i own this cd', 'own it', 'love at first listen', 'underrated',
                       'fucking awesome', 'in my collection', 'check',
                       'my private work station', 'top cd', 'fav', 'love it', 'happy', 'owned', 'mandatory', 'the best',
                       'streamable', 'collected', 'reviewed in the guardian',
                       'to check out', 'overrated', 'to buy', 'i own', '1', 'favs', 'essential', '5', 'music',
                       'my music', 'music to download', ]:
        return None

    for substring in ['artist', 'bpm', 'best of', 'album', 'favorite', 'favourite', 'vinyl', 'seen ', 'albun', 'need']:
        if substring in tag:
            return None

    tag_reverse_map = {
        'fusion': {
            'crossover'
        },

        'drum and bass': {
            'drum \'n bass',
            'drum \'n\' bass',
            'drum & bass',
            'drum n bass',
            'drum n\' bass',
            'drum\'n bass',
            'drum\'n\'bass',
            'drum&bass',
            'drumandbass',
            'drumm and bass',
            'drumm n bass',
            'drumn and bass',
            'drumnbass',
            'drums and base',
            'drums and bass',
            'dnb',
            'd&b',
        },

        'hip hop': {
            'hip-hop',
            'hiphop',
        },

        'trip-hop': {
            'triphop',
            'trip hop',
        }
    }

    tag_remap = {}
    for dest_tag in tag_reverse_map:
        for curr_tag in tag_reverse_map[dest_tag]:
            tag_remap[curr_tag] = dest_tag

    if tag in tag_remap:
        tag = tag_remap[tag]

    if capitalize:
        tag = capitalize_tag(tag)

        tag_allcaps = ["av", "uk"]

        if tag.lower() in tag_allcaps:
            tag = tag.upper()

    return tag


def tag_filter_all(tags: Dict[str, int], ignore_substrings: List[str], capitalize=False):
    accepted_tags = OrderedDict()
    for tag in tags:
        curr_tag = tag_filter(tag, ignore_substrings, capitalize)

        # If the tag was filtered, mark as ignore
        if not curr_tag:
            continue

        # If a remapped tag clashes with another, and its score is lower, mark as ignore
        if curr_tag in accepted_tags and accepted_tags[curr_tag] > tags[tag]:
            continue
        accepted_tags[curr_tag] = tags[tag]

    return accepted_tags


non_capitalized = ["a", "an", "the", "and", "but", "or", "for", "nor", "in", "to", "at" "on", "by", "from", "nor",
                   "yet", "so"]


def capitalize_tag(tag_str):
    words = [x.capitalize() if x not in non_capitalized else x for x in tag_str.split(" ")]
    if len(words):
        words[0] = words[0].capitalize()

    return " ".join(words)


def extract_track_disc(filename):
    track = None
    disc = None

    match = re.match(r"^(\d{2,4})", filename)
    if match:
        track = int(match[0][-2:])
    if match and len(match[0]) > 2:
        disc = int(match[0][:-2])

    if not track:
        match = re.findall(r"[ \-_]{1}(\d{1,2})[ \-_]", filename)
        if match and len(match) == 1:
            track = int(match[0])

    return track, disc


def extract_release_year(folder_name: Optional[int]):
    matches = [int(x[1]) for x in re.findall(r"(^|-|_| )(\d{4})($|-|_| )", folder_name)]
    if len(matches) == 1 and matches[0] >= 1920 and matches[0] <= datetime.date.today().year:
        return matches[0]

    matches = [int(x[1]) for x in re.findall(r"(\(|\[|{)(\d{4})(\)|\]|})", folder_name)]
    if len(matches) == 1 and matches[0] >= 1920 and matches[0] <= datetime.date.today().year:
        return matches[0]


from metafix.constants import ViolationType


class Violation:
    def __init__(self, violation_type: ViolationType, message: str) -> None:
        self.violation_type = violation_type
        self.message = message

    def __repr__(self) -> str:
        return self.message



import copy
import logging
import os
import re
import time
from collections import OrderedDict
from typing import List

from lastfmcache import LastfmCache
from ordered_set import OrderedSet

from lastfmcache.lastfmcache import LastfmRelease
from metafix.Release import Release
from metafix.Violation import Violation
from metafix.constants import ViolationType, upgrade_message, ReleaseSource, ReleaseCategory
from metafix.functions import flatten_artists, normalize_track_title, split_release_title, tag_filter_all, \
    extract_track_disc, unique, extract_release_year, normalize_artist_name, normalize_release_title


class ReleaseValidator:

    def __init__(self, lastfm: LastfmCache = None):
        self.lastfm = lastfm
        self.forbidden_comment_substrings = set()
        self.lastfm_track_title_validation = True

    def disable_lastfm_track_title_validation(self) -> None:
        self.lastfm_track_title_validation = False

    def add_forbidden_comment_substring(self, forbidden_comment_substring: str) -> None:
        self.forbidden_comment_substrings.add(forbidden_comment_substring.lower())

    def validate(self, release: Release) -> List[Violation]:
        violations = OrderedSet()

        # leading/trailing whitespace
        for filename, track in release.tracks.items():
            if track.artists != track.strip_whitespace_artists():
                violations.add(
                    Violation(ViolationType.ARTIST_WHITESPACE,
                              "File '{0}' has leading/trailing whitespace in its Artist(s)".format(filename)))

        for filename, track in release.tracks.items():
            if track.release_artists != track.strip_whitespace_release_artists():
                violations.add(
                    Violation(ViolationType.RELEASE_ARTIST_WHITESPACE,
                              "File '{0}' has leading/trailing whitespace in its Album/Release Artist(s)"
                              .format(filename)))

        for filename, track in release.tracks.items():
            if track.date != track.strip_whitespace_date():
                violations.add(
                    Violation(ViolationType.DATE_WHITESPACE,
                              "File '{0}' has leading/trailing whitespace in its Year/Date".format(filename)))

        for filename, track in release.tracks.items():
            if track.release_title != track.strip_whitespace_release_title():
                violations.add(
                    Violation(ViolationType.RELEASE_TITLE_WHITESPACE,
                              "File '{0}' has leading/trailing whitespace in its Album/Release Title"
                              .format(filename)))

        for filename, track in release.tracks.items():
            if track.track_title != track.strip_whitespace_track_title():
                violations.add(
                    Violation(ViolationType.TRACK_TITLE_WHITESPACE,
                              "File '{0}' has leading/trailing whitespace in its Track Title"
                              .format(filename)))

        for filename, track in release.tracks.items():
            if track.genres != track.strip_whitespace_genres():
                violations.add(
                    Violation(ViolationType.GENRE_WHITESPACE,
                              "File '{0}' has leading/trailing whitespace in its Genre(s)"
                              .format(filename)))

        # release date
        if not release.validate_release_date():
            violations.add(
                Violation(ViolationType.DATE_INCONSISTENT, "Release contains blank or inconsistent 'Date' tags"))

        # artists
        if release.blank_artists():
            violations.add(
                Violation(ViolationType.ARTIST_BLANK,
                          "Release contains {0} tracks with missing 'Artist' tags".format(release.blank_artists())))

        # track titles
        if release.blank_track_titles():
            violations.add(
                Violation(ViolationType.TRACK_TITLE_BLANK, "Release contains {0} tracks with missing 'Track Title' tags"
                          .format(release.blank_track_titles())))

        # release artist
        release_artists = release.validate_release_artists()
        if not release_artists:
            violations.add(
                Violation(ViolationType.RELEASE_ARTIST_INCONSISTENT,
                          "Release contains blank or inconsistent 'Album/Release Artist' tags"))

        # if the lastfmcache is present, validate the release artist
        validated_release_artists = release_artists
        if self.lastfm and len(release_artists) == 1:
            validated_release_artists = []
            for artist in release_artists:
                try:
                    validated_release_artist = self.lastfm.get_artist(artist.strip()).artist_name

                    if validated_release_artist != artist:
                        violations.add(
                            Violation(ViolationType.RELEASE_ARTIST_SPELLING,
                                      "Incorrectly spelled Album/Release Artist '{0}' (should be '{1}')".format(
                                          artist, validated_release_artist)))

                    validated_release_artists.append(validated_release_artist)
                except LastfmCache.ArtistNotFoundError:
                    violations.add(
                        Violation(ViolationType.ARTIST_LOOKUP, "Lookup failed of release artist '{release_artist}'"
                                  .format(release_artist=artist.strip())))

        # release title
        release_title = release.validate_release_title()
        if not release_title:
            violations.add(Violation(ViolationType.RELEASE_TITLE_INCONSISTENT,
                                     "Release contains blank or inconsistent 'Album/Release Title' tags"))

        bracket_pairs = [["[", "]"], ["(", ")"], ["{", "}"]]

        if release_title:
            # check if "[Source]" is contained in the release title
            for source in ReleaseSource:
                for brackets in bracket_pairs:
                    curr_source = "{0}{1}{2}".format(brackets[0], source.value, brackets[1])
                    if curr_source.lower() in release_title.lower():
                        violations.add(Violation(ViolationType.RELEASE_TITLE_SOURCE,
                                                 "Release title contains source {0}".format(curr_source)))

            # check if the release title ends with a space and a source name, without brackets
            for source in [x for x in ReleaseSource]:
                if release_title.lower().endswith(" {0}".format(source.value.lower())):
                    violations.add(Violation(ViolationType.RELEASE_TITLE_SOURCE,
                                             "Release title ends with source {0}".format(source.value)))

            # check if "[Category]" is contained in the release title
            for category in ReleaseCategory:
                for brackets in bracket_pairs:
                    curr_category = "{0}{1}{2}".format(brackets[0], category.value, brackets[1])
                    if curr_category.lower() in release_title.lower():
                        violations.add(Violation(ViolationType.RELEASE_TITLE_CATEGORY,
                                                 "Release title contains category {0}".format(curr_category)))

            # check if the release title ends with a space and a category name, without brackets (except Album)
            for category in [x for x in ReleaseCategory if x is not ReleaseCategory.ALBUM]:
                if release_title.lower().endswith(" {0}".format(category.value.lower())):
                    violations.add(Violation(ViolationType.RELEASE_TITLE_CATEGORY,
                                             "Release title ends with category {0}".format(category.value)))

        # lastfm artist validations
        if self.lastfm and release_title and len(validated_release_artists):
            # extract (edition info) from release titles
            release_title, _ = split_release_title(normalize_release_title(release_title))

            flattened_artist = flatten_artists(validated_release_artists)
            lastfm_release = None

            try:
                lastfm_release = self.lastfm.get_release(flattened_artist, release_title)
            except LastfmCache.ReleaseNotFoundError as e:
                logging.getLogger(__name__).error(e)

            if lastfm_release:
                # release title
                if lastfm_release.release_name != release_title and \
                        ReleaseValidator.__lastfm_can_fix_release_title(release_title, lastfm_release.release_name):
                    violations.add(
                        Violation(ViolationType.RELEASE_TITLE_SPELLING,
                                  "Incorrectly spelled Album/Release name '{0}' (should be '{1}')"
                                  .format(release_title, lastfm_release.release_name)))

                # dates
                if lastfm_release.release_date:
                    date = next(iter(release.tracks.values())).date
                    if lastfm_release.release_date != date and \
                            (not date or len(lastfm_release.release_date) >= len(date)):
                        violations.add(
                            Violation(ViolationType.DATE_INCORRECT, "Incorrect Release Date '{0}' (should be '{1}')"
                                      .format(date, lastfm_release.release_date)))

                # tags/genres (only fail if 0-1 genres - i.e. lastfm tags have never been applied)
                release_genres = release.validate_genres()
                lastfm_tags = self.__get_lastfm_tags(release_title, validated_release_artists)
                if len(release_genres) < 2 <= len(lastfm_tags):
                    violations.add(
                        Violation(ViolationType.BAD_GENRES, "Bad release genres: [{0}] (should be [{1}])"
                                  .format(", ".join(release_genres), ", ".join(lastfm_tags))))

                # match and validate track titles (intersection only)
                if self.lastfm_track_title_validation:
                    for track in release.tracks.values():
                        if track.track_number in lastfm_release.tracks:
                            lastfm_title = normalize_track_title(lastfm_release.tracks[track.track_number].track_name)
                            if not track.track_title or track.track_title.lower() != lastfm_title.lower():
                                violations.add(
                                    Violation(ViolationType.INCORRECT_TRACK_TITLE,
                                              "Incorrect track title '{0}' should be: '{1}'".format(track.track_title,
                                                                                                    lastfm_title)))

            # track artists
            for track in release.tracks.values():
                for artist in track.artists:
                    while True:
                        try:
                            validated_artist = self.lastfm.get_artist(normalize_artist_name(artist)).artist_name
                            if validated_artist != artist:
                                violations.add(
                                    Violation(ViolationType.TRACK_ARTIST_SPELLING,
                                              "Incorrectly spelled Track Artist '{0}' (should be '{1}')"
                                              .format(artist, validated_artist)))
                            break
                        except LastfmCache.ArtistNotFoundError:  # as e:
                            # violations.add(str(e))
                            break
                        except LastfmCache.LastfmCacheError:
                            time.sleep(1)

            # release artists
            for track in release.tracks.values():
                for artist in track.release_artists:
                    while True:
                        try:
                            validated_artist = self.lastfm.get_artist(normalize_artist_name(artist)).artist_name
                            if validated_artist != artist:
                                violations.add(
                                    Violation(ViolationType.RELEASE_ARTIST_SPELLING,
                                              "Incorrectly spelled Release Artist '{0}' (should be '{1}')"
                                              .format(artist, validated_artist)))
                            break
                        except LastfmCache.ArtistNotFoundError:  # as e:
                            # violations.add(str(e))
                            break
                        except LastfmCache.LastfmCacheError:
                            time.sleep(1)

        validated_track_numbers = release.validate_track_numbers()
        if validated_track_numbers:
            flattened_track_nums = []
            for disc in validated_track_numbers:
                flattened_track_nums.append(
                    "\nDisc " + str(disc) + ": " + ",".join(str(i) for i in validated_track_numbers[disc]))
            violations.add(
                Violation(ViolationType.MISSING_TRACKS,
                          "Release does not have a full set of tracks:{0}".format("".join(flattened_track_nums))))

        validated_total_tracks = release.validate_total_tracks()
        for disc in validated_total_tracks:
            violations.add(
                Violation(ViolationType.TOTAL_TRACKS_INCONSISTENT,
                          "Release disc {0} has blank, inconsistent or incorrect 'Total Tracks' tags".format(disc)))

        # disc number
        validated_disc_numbers = release.validate_disc_numbers()
        if validated_disc_numbers:
            violations.add(
                Violation(ViolationType.MISSING_DISCS, "Release does not have a full set of discs: {0}"
                          .format(", ".join(str(i) for i in validated_disc_numbers))))

        # total discs
        if not release.validate_total_discs():
            violations.add(
                Violation(ViolationType.TOTAL_DISCS_INCONSISTENT, "Release has incorrect 'Total Discs' tags"))

        # file type
        if len(release.get_tag_types()) != 1:
            violations.add(
                Violation(ViolationType.TAG_TYPES_INCONSISTENT, "Release has inconsistent tag types: {0}"
                          .format(", ".join([str(x) for x in release.get_tag_types()]))))

        # bitrate - CBR/VBR/Vx/APS/APE
        if len(release.get_codecs()) != 1:
            violations.add(
                Violation(ViolationType.CODECS_INCONSISTENT,
                          "Release has inconsistent codecs: [{0}]".format(", ".join(release.get_codecs()))))

        if len(unique([int(x / 1000) for x in release.get_cbr_bitrates()])) > 1:
            violations.add(
                Violation(ViolationType.CBR_INCONSISTENT, "Release has inconsistent CBR bitrates: {0}"
                          .format(", ".join([str(x) for x in release.get_cbr_bitrates()]))))

        # track titles
        for filename in release.tracks:
            correct_filename = release.tracks[filename].get_filename(release.is_va())
            if correct_filename and filename != correct_filename:
                violations.add(
                    Violation(ViolationType.FILENAME,
                              "Invalid filename: {0} - should be '{1}'".format(filename, correct_filename)))

        # forbidden comment substrings
        for track in release.tracks.values():
            if not track.comment:
                continue
            for substr in self.forbidden_comment_substrings:
                if substr in track.comment.lower():
                    violations.add(
                        Violation(ViolationType.COMMENT_SUBSTRING,
                                  "Invalid comment: contains forbidden substring '{0}'".format(substr)))

        release.num_violations = len(violations)

        return list(violations)

    def fix(self, release_in: Release, folder_name: str) -> Release:
        release = copy.deepcopy(release_in)

        # fix leading/trailing whitespace
        self.__fix_whitespace(release)

        # extract missing track and disc numbers from filenames
        self.__extract_track_disc_numbers_from_filenames(release)

        # extract missing year from folder name
        self.__extract_year_from_folder_name(release, folder_name)

        # extract missing disc numbers from folder name
        self.__extract_disc_numbers_from_folder_name(release)

        # normalize track titles
        self.__normalize_track_titles(release)

        # normalize release artists
        release_artists = self.__normalize_release_artists(release)

        # fix release artists using last.fm
        validated_release_artists = self.__lastfm_release_artists(release, release_artists)

        # fix release title
        self.__fix_release_title(release)

        # lastfm fixes
        self.__lastfm_fixes(release, validated_release_artists)

        # reorder track numbers if they are sequential across discs
        release.resequence_track_numbers()

        # fix missing total track numbers
        self.__fix_missing_total_tracks(release)

        # if disc number is missing and there appears to only be one disc, set to 1
        self.__disc_number_best_guess(release)

        # fix missing total disc numbers
        self.__fix_missing_total_discs(release)
        
        # fix forbidden comment substrings
        self.__fix_forbidden_comment_substrings(release)

        return release

    # tags/genres
    def __get_lastfm_tags(self, release_title: str, release_artists: List[str]):
        flattened_artist = flatten_artists(release_artists)
        lastfm_release = self.lastfm.get_release(flattened_artist, release_title)

        lastfm_tags = [x for x in tag_filter_all(lastfm_release.tags, release_artists + [release_title], True)]
        if not lastfm_tags:
            filtered_artists = [x for x in release_artists
                                if x.lower() not in ["various artist", "various artists", "va"]]
            artists_tags = [self.lastfm.get_artist(artist).tags for artist in filtered_artists]
            weighted_tags = OrderedDict()
            for artist_tags in artists_tags:
                for tag in artist_tags:
                    if tag not in weighted_tags or (tag in weighted_tags and weighted_tags[tag] < artist_tags[tag]):
                        weighted_tags[tag] = artist_tags[tag]
            lastfm_tags = [x for x in tag_filter_all(weighted_tags, filtered_artists + [release_title], True)]

        return [x for x in lastfm_tags]

    @staticmethod
    def __fix_whitespace(release: Release) -> None:
        """fix leading/trailing whitespace"""
        for track in release.tracks.values():
            track.artists = track.strip_whitespace_artists()
            track.release_artists = track.strip_whitespace_release_artists()
            track.date = track.strip_whitespace_date()
            track.release_title = track.strip_whitespace_release_title()
            track.track_title = track.strip_whitespace_track_title()
            track.genres = track.strip_whitespace_genres()

    @staticmethod
    def __extract_track_disc_numbers_from_filenames(release: Release) -> None:
        """extract missing track and disc numbers from filenames"""
        validated_track_numbers = release.validate_track_numbers()
        validated_disc_numbers = release.validate_disc_numbers()
        for path in release.tracks:
            track_num, disc_num = extract_track_disc(os.path.split(path)[-1])
            if (not release.tracks[path].track_number and track_num) or validated_track_numbers:
                release.tracks[path].track_number = track_num
                if validated_track_numbers and disc_num:
                    release.tracks[path].disc_number = disc_num
            if (not release.tracks[path].disc_number and disc_num) or validated_disc_numbers:
                release.tracks[path].disc_number = disc_num

    @staticmethod
    def __extract_year_from_folder_name(release: Release, folder_name: str) -> None:
        """extract missing year from folder name"""
        extracted_year = extract_release_year(folder_name)
        if not release.validate_release_date() and extracted_year:
            for track in release.tracks.values():
                track.date = str(extracted_year)

    @staticmethod
    def __extract_disc_numbers_from_folder_name(release: Release) -> None:
        """extract missing disc numbers from folder name"""
        for filename in release.tracks:
            if not release.tracks[filename].disc_number:
                match = re.findall(r'(?i)(disc|disk|cd|part) ?(\d{1,2})', os.path.split(filename)[0])
                if match:
                    release.tracks[filename].disc_number = int(match[0][1])

    @staticmethod
    def __normalize_track_titles(release: Release) -> None:
        """normalize track titles"""
        for track in release.tracks.values():
            if track.track_title:
                normalized_title = normalize_track_title(track.track_title)
                if track.track_title.lower() != normalized_title.lower():
                    track.track_title = normalized_title

    @staticmethod
    def __normalize_release_title(release: Release) -> None:
        """normalize release title"""
        release_title = release.validate_release_title()
        for track in release.tracks.values():
            if track.release_title != release_title:
                track.release_title = release_title

    @staticmethod
    def __normalize_release_artists(release: Release) -> List[str]:
        """normalize release artists"""
        release_artists = release.validate_release_artists()
        if not release_artists:
            release_artists = release.extract_release_artist()

        return release_artists

    def __lastfm_release_artists(self, release: Release, release_artists: List[str]) -> List[str]:
        """fix release artists using last.fm"""
        if not self.lastfm or not release_artists:
            return []

        validated_release_artists = []
        for artist in release_artists:
            while True:
                try:
                    validated_release_artists.append(self.lastfm.get_artist(artist).artist_name)
                    break
                except LastfmCache.ArtistNotFoundError:
                    break
                except LastfmCache.UpgradeRequiredError:
                    logging.getLogger(__name__).error(upgrade_message)
                    exit(1)
                except LastfmCache.ConnectionError:
                    logging.getLogger(__name__).error("Connection error while retrieving artist, retrying...")
                    time.sleep(1)
                except LastfmCache.LastfmCacheError:
                    logging.getLogger(__name__).error("Server error while retrieving artist, retrying...")
                    time.sleep(1)

        # update release artists for all tracks, if all were validated
        if len(validated_release_artists) == len(release_artists):
            for track in release.tracks.values():
                track.release_artists = validated_release_artists

        return validated_release_artists

    @staticmethod
    def __fix_release_title(release: Release) -> None:
        """fix release title"""
        release_title = release.validate_release_title()
        if not release_title:
            return

        for category_stub in ["CD", "CDS", "CDM"]:
            if release_title.endswith(" {0}".format(category_stub)):
                release_title = release_title[:-(len(category_stub)+1)]


        bracket_pairs = [["[", "]"], ["(", ")"], ["{", "}"]]

        # Remove "[Source]" from release title
        for source in ReleaseSource:
            for brackets in bracket_pairs:
                curr_source = "{0}{1}{2}".format(brackets[0], source.value, brackets[1])
                pattern = "(?i)( )?" + re.escape(curr_source)
                if re.search(pattern, release_title):
                    release_title = re.sub(pattern, "", release_title)

        # Remove trailing ' Source' from release title
        for source in [x for x in ReleaseSource]:
            pattern = "(?i)( \-)? " + source.value + "$"
            if re.search(pattern, release_title):
                release_title = re.sub(pattern, "", release_title)

        # Remove "[Category]" from release title
        for category in ReleaseCategory:
            for brackets in bracket_pairs:
                curr_category = "{0}{1}{2}".format(brackets[0], category.value, brackets[1])
                pattern = "(?i)( )?" + re.escape(curr_category)
                if re.search(pattern, release_title):
                    release_title = re.sub(pattern, "", release_title)

        # Remove trailing ' category' from release title
        for category in [x for x in ReleaseCategory if x is not ReleaseCategory.ALBUM]:
            pattern = "(?i)( \-)? " + category.value + "$"
            if re.search(pattern, release_title):
                release_title = re.sub(pattern, "", release_title)

        # Overwrite the release's 'release title' tags
        for track in release.tracks.values():
            if track.release_title != release_title:
                track.release_title = release_title

    @staticmethod
    def __fix_missing_total_tracks(release: Release) -> None:
        # fix missing total track numbers
        validated_disc_numbers = release.get_total_tracks()
        for track in release.tracks.values():
            disc_number = track.disc_number if track.disc_number else 1
            if validated_disc_numbers.get(disc_number):
                track.total_tracks = validated_disc_numbers[disc_number]

    @staticmethod
    def __disc_number_best_guess(release: Release) -> None:
        """if disc number is missing and there appears to only be one disc, set to """
        validated_disc_numbers = release.get_total_tracks()
        if not release.validate_total_discs() and len(validated_disc_numbers) == 1:
            for track in release.tracks.values():
                track.disc_number = 1

    @staticmethod
    def __fix_missing_total_discs(release: Release) -> None:
        # fix missing total disc numbers
        if not release.validate_total_discs():
            total_discs = release.get_total_discs()
            if total_discs:
                for track in release.tracks.values():
                    track.total_discs = total_discs

    def __fix_forbidden_comment_substrings(self, release: Release) -> None:
        # forbidden comment substrings
        for track in release.tracks.values():
            if track.comment:
                for substr in self.forbidden_comment_substrings:
                    if substr in track.comment.lower():
                        track.comment = None

    def __lastfm_fixes(self, release: Release, release_artists: List[str]) -> None:
        # lastfm fixes

        release_title = release.validate_release_title()

        if not self.lastfm or not len(release_artists) or not release_title:
            return

        # extract (edition info) from release titles
        release_title, release_edition = split_release_title(normalize_release_title(release_title))

        flattened_artist = flatten_artists(release_artists)

        lastfm_release = None

        while True:
            try:
                lastfm_release = self.lastfm.get_release(flattened_artist, release_title)
                break
            except LastfmCache.ReleaseNotFoundError as e:
                logging.getLogger(__name__).error(e)
                break
            except LastfmCache.UpgradeRequiredError:
                logging.getLogger(__name__).error(upgrade_message)
                exit(1)
            except LastfmCache.ConnectionError:
                logging.getLogger(__name__).error("Connection error while retrieving release, retrying...")
                time.sleep(1)
            except LastfmCache.LastfmCacheError:
                logging.getLogger(__name__).error("Server error while retrieving release, retrying...")
                time.sleep(1)

        self.__lastfm_release_fixes(release, lastfm_release, release_artists, release_title, release_edition)

        # fix track artists using lastfm
        self.__lastfm_fix_track_artists(release)

        # fix release artists using lastfm
        self.__lastfm_fix_release_artists(release)

    def __lastfm_release_fixes(self, release: Release, lastfm_release: LastfmRelease, release_artists: List[str],
                               release_title: str, release_edition: str) -> None:
        """lastfm release fixes"""

        if release_artists:
            for track in release.tracks.values():
                track.release_artists = release_artists

        if not lastfm_release:
            return

        # release title
        if lastfm_release.release_name != release_title and \
                ReleaseValidator.__lastfm_can_fix_release_title(release_title, lastfm_release.release_name):
            release_title_full = lastfm_release.release_name
            if release_edition:
                release_title_full = "{0} {1}".format(lastfm_release.release_name, release_edition)

            for track in release.tracks.values():
                track.release_title = release_title_full

        # dates
        if lastfm_release.release_date:
            for track in release.tracks.values():
                if lastfm_release.release_date and lastfm_release.release_date != track.date:
                    track.date = lastfm_release.release_date

        # tags/genres (only fail if 0-1 genres - i.e. lastfm tags have never been applied)
        release_genres = release.validate_genres()
        lastfm_tags = self.__get_lastfm_tags(release_title, release_artists)
        if len(release_genres) < 2 <= len(lastfm_tags):
            for track in release.tracks.values():
                track.genres = lastfm_tags

        # fill missing track numbers from lastfm
        for track in release.tracks.values():
            if track.track_number:
                continue

            track_num_matches = [int(x) for x in lastfm_release.tracks
                                 if normalize_track_title(lastfm_release.tracks[x].track_name).lower() ==
                                 normalize_track_title(track.track_title).lower()]
            if track_num_matches and len(track_num_matches) == 1 and not \
                    [x.track_number for x in release.tracks.values() if x.track_number == track_num_matches[0]]:
                track.track_number = track_num_matches[0]

        # match and validate track titles (intersection only)
        track_numbers_validated = not release.validate_track_numbers()
        for track in release.tracks.values():
            if track.track_number in lastfm_release.tracks:
                lastfm_title = normalize_track_title(lastfm_release.tracks[track.track_number].track_name)

                if track.track_title != lastfm_title:
                    # if the track title is missing, or if it is lowercase and there is a case insensitive match
                    if (not track.track_title and track_numbers_validated) or \
                            (track.track_title.islower() and track.track_title.lower() == lastfm_title.lower()):
                        track.track_title = lastfm_title

                    # case insensitive match, tag version has no capital letters
                    elif track.track_title.lower() == lastfm_title.lower() \
                            and track.track_title.lower() == track.track_title:
                        track.track_title = lastfm_title


    def __lastfm_fix_track_artists(self, release: Release) -> None:
        """fix track artists using lastfm"""
        for track in release.tracks.values():
            validated_artists = []
            for artist in track.artists:
                while True:
                    try:
                        validated_artists.append(self.lastfm.get_artist(normalize_artist_name(artist)).artist_name)
                        break
                    except LastfmCache.ArtistNotFoundError:
                        validated_artists.append(artist)
                        break
                    except LastfmCache.UpgradeRequiredError:
                        logging.getLogger(__name__).error(upgrade_message)
                        exit(1)
                    except LastfmCache.ConnectionError:
                        logging.getLogger(__name__).error("Connection error while retrieving artist, retrying...")
                        time.sleep(1)
                    except LastfmCache.LastfmCacheError:
                        logging.getLogger(__name__).error("Server error while retrieving artist, retrying...")
                        time.sleep(1)
                    except AttributeError:  # TODO remove when pylast is fixed
                        break

            track.artists = validated_artists

    def __lastfm_fix_release_artists(self, release: Release) -> None:
        """fix release artists using lastfm"""
        for track in release.tracks.values():
            validated_artists = []
            for artist in track.release_artists:
                while True:
                    try:
                        validated_artists.append(self.lastfm.get_artist(normalize_artist_name(artist)).artist_name)
                        break
                    except LastfmCache.ArtistNotFoundError:
                        break
                    except LastfmCache.UpgradeRequiredError:
                        logging.getLogger(__name__).error(upgrade_message)
                        exit(1)
                    except LastfmCache.ConnectionError:
                        logging.getLogger(__name__).error("Connection error while retrieving artist, retrying...")
                        time.sleep(1)
                    except LastfmCache.LastfmCacheError:
                        logging.getLogger(__name__).error("Server error while retrieving artist, retrying...")
                        time.sleep(1)
                    except AttributeError:  # TODO remove when pylast is fixed
                        break

            if len(validated_artists) == len(track.artists):
                track.release_artists = validated_artists

    @staticmethod
    def __lastfm_can_fix_release_title(release_title: str, fixed_title: str) -> bool:
        if fixed_title.islower() and not release_title.islower():
            return False

        exception_substrings = {'fabric'}
        for substr in exception_substrings:
            if substr.lower() in release_title.lower():
                return False

        return True



from typing import List

from cleartag import Track as cleartagTrack
from metafix.functions import unique


class Track(cleartagTrack):

    def strip_whitespace_artists(self) -> List[str]:
        return unique([x.strip() for x in self.artists])

    def strip_whitespace_release_artists(self) -> List[str]:
        return unique([x.strip() for x in self.release_artists])

    def strip_whitespace_date(self) -> str:
        return None if not self.date else self.date.strip()

    def strip_whitespace_release_title(self) -> str:
        return None if not self.release_title else self.release_title.strip()

    def strip_whitespace_track_title(self) -> str:
        return None if not self.track_title else self.track_title.strip()

    def strip_whitespace_genres(self) -> List[str]:
        return unique([x.strip() for x in self.genres])



from __future__ import annotations

import os
from collections import OrderedDict
from typing import Dict, List, Optional

from cleartag.enums.TagType import TagType
from cleartag.functions import normalize_path_chars
from metafix.Track import Track
from metafix.constants import ReleaseCategory, ReleaseSource
from metafix.functions import unique, flatten_artists, get_category_fix_name, normalize_release_title


class Release:

    def __init__(self, tracks: Dict[str, Track], manual_category: ReleaseCategory = ReleaseCategory.UNKNOWN,
                 manual_source = ReleaseSource.UNKNOWN):
        self.tracks = tracks
        self.category = manual_category
        self.source = manual_source
        self.num_violations = None
        self.guess_category()
        self.sort()

    def __eq__(self, other):
        return self.tracks == other.tracks

    def __repr__(self):
        track0 = self.tracks[next(iter(self.tracks))]
        date = "<date not found>" if not track0.date else track0.date.split("-")[0]
        return "{0} - {1} - {2}"\
            .format(flatten_artists(track0.release_artists), date, track0.release_title)

    def sort(self):
        reordered = OrderedDict()

        while len(self.tracks):
            smallest_filename = next(iter(self.tracks))
            smallest_track = self.tracks[smallest_filename]
            for filename in self.tracks:
                if self.tracks[filename] < smallest_track:
                    smallest_filename = filename
                    smallest_track = self.tracks[filename]
            del self.tracks[smallest_filename]
            reordered[smallest_filename] = smallest_track

        self.tracks = reordered


    def guess_category(self) -> None:
        if self.category:
            return

        # clean release name, and category
        _, self.category = get_category_fix_name(self)

        # extract additional artists
        additional_artists = []
        for track in self.tracks.values():
            for artist in track.artists:
                if artist not in self.tracks[next(iter(self.tracks))].release_artists:
                    additional_artists.append(artist)

        if len(self.tracks) < 4 and len(unique(additional_artists)) == 1:
            self.category = ReleaseCategory.SINGLE
        elif len(self.tracks) < 6 and len(unique(additional_artists)) == 1:
            self.category = ReleaseCategory.EP
        elif len(unique(additional_artists)) > len(self.tracks) / 2:
            self.category = ReleaseCategory.COMPILATION
        else:
            self.category = ReleaseCategory.ALBUM

    def is_va(self):
        return self.category in {ReleaseCategory.COMPILATION, ReleaseCategory.VIDEO_GAME_MUSIC, ReleaseCategory.MIX,
                                 ReleaseCategory.MIXTAPE, ReleaseCategory.SOUNDTRACK}

    def get_release_codec_setting(self, short=True) -> str:

        # check for mismatched tag types
        tag_types = unique([track.stream_info.tag_type for track in self.tracks.values()])
        if len(tag_types) != 1:
            return ""

        # check for mismatched codec settings
        codec_settings = unique([track.get_codec_setting_str(short=True) for track in self.tracks.values()])
        if len(codec_settings) != 1:
            return ""

        prefix_str = ""
        if not short:
            if tag_types[0] == TagType.ID3:
                prefix_str = "MP3 "
            elif tag_types[0] == TagType.MP4:
                prefix_str = "MP4 "

        # check for mismatched CBR bitrates
        if codec_settings[0] == "CBR":
            cbr_bitrates = unique([round(track.stream_info.bitrate, -3) for track in self.tracks.values()])
            if len(cbr_bitrates) != 1:
                return ""

            cbr_bitrate = int(round(cbr_bitrates[0] / 1000))
            cbr_bitrate += cbr_bitrate % 2

            return "{0}CBR{1}".format(prefix_str, cbr_bitrate)

        elif codec_settings[0] == "VBR":
            average_bitrate = round(
                sum([track.stream_info.bitrate * track.stream_info.length for track in self.tracks.values()]) /
                sum([track.stream_info.length for track in self.tracks.values()]) / 1000)

            return "{0}VBR{1}".format(prefix_str, average_bitrate)

        else:
            return "{0}{1}".format(prefix_str, codec_settings[0])

    def get_codec_rank(self) -> int:
        """Return a codec indicating the release's codec ranking. Higher is better."""
        return Release.get_codec_rank_static(self.get_release_codec_setting())

    @staticmethod
    def get_codec_rank_static(codec: str) -> int:
        """Return a codec indicating the release's codec ranking. Higher is better."""
        if not codec:
            return 0

        rank = 1000

        if codec in ["FLAC", "24bit FLAC", "V0"]: return rank
        rank -= 1
        if codec == "APE": return rank
        rank -= 1
        if codec == "CBR320": return rank
        rank -= 1
        if codec == "V1": return rank
        rank -= 1
        if codec == "vbr-old V1": return rank
        rank -= 1
        if codec == "V2": return rank
        rank -= 1
        if codec == "APS": return rank
        rank -= 1
        if codec == "V3": return rank
        rank -= 1
        if codec == "vbr-old V3": return rank
        rank -= 1
        if codec == "V4": return rank
        rank -= 1
        if codec == "APM": return rank
        rank -= 1

        # VBR/CBR/ABR: 320k > bitrate >= 256k
        if len(codec) >= 5 and codec[:3] in ["VBR", "CBR", "ABR"] and codec[3:].isdigit():
            curr_bitrate = 320
            bitrate = int(codec[3:])

            while curr_bitrate >= 256:
                for curr_setting in ["VBR", "CBR", "ABR"]:
                    if curr_setting == codec[:3] and bitrate <= curr_bitrate and bitrate >= curr_bitrate - 64:
                        return rank
                    rank -= 1

                curr_bitrate -= 64

        # V5 -> V9
        for lame in ["V", "vbr-old V"]:
            for num in range(5, 10):
                if codec == "{0}{1}".format(lame, num):
                    return rank
                rank -= 1

        # VBR/CBR/ABR: 255k > bitrate >= 0
        if len(codec) >= 5 and codec[:3] in ["VBR", "CBR", "ABR"] and codec[3:].isdigit():
            curr_bitrate = 192
            bitrate = int(codec[3:])

            while curr_bitrate > 0:
                for curr_setting in ["VBR", "CBR", "ABR"]:
                    if curr_setting == codec[:3] and bitrate <= curr_bitrate and bitrate >= curr_bitrate - 64:
                        return rank
                    rank -= 1

                curr_bitrate -= 64

        if "lame vbr method" in codec: return rank
        rank -= 1

        return -1


    def get_folder_name(self, codec_short: bool = True, group_by_category: bool = False):

        assert self.validate_release_date(), "Release date validation failed"
        assert self.validate_release_title(), "Release title validation failed"
        assert self.validate_release_artists(), "Release artists validation failed"
        assert self.validate_codec(), "Codec validation failed"

        track1 = self.tracks[next(iter(self.tracks))]

        # clean release name, and category
        release_name, _ = get_category_fix_name(self)

        release_artist = flatten_artists(track1.release_artists)
        year = track1.date.split("-")[0]

        release_category_str = "[{0}] ".format(self.category.value) if self.category != ReleaseCategory.ALBUM else ""
        release_source_str = "[{0}] ".format(self.source.value) if self.source != ReleaseSource.CD else ""

        title_first_categories = {ReleaseCategory.COMPILATION, ReleaseCategory.MIX, ReleaseCategory.MIXTAPE,
                                  ReleaseCategory.VIDEO_GAME_MUSIC, ReleaseCategory.SOUNDTRACK}

        # folder name
        if group_by_category is False and self.category in title_first_categories:
            return normalize_path_chars(
                "VA - {release_name} - {year} - {release_artist} "
                "{release_category_str}{release_source_str}[{release_codec}]"
                .format(release_name=release_name,
                        year=year,
                        release_artist=release_artist,
                        release_category_str=release_category_str,
                        release_codec=self.get_release_codec_setting(codec_short),
                        release_source_str=release_source_str))

        elif self.category in title_first_categories:
            return normalize_path_chars(
                "{release_name} - {year} - {release_artist} "
                "{release_category_str}{release_source_str}[{release_codec}]"
                .format(release_name=release_name,
                        year=year,
                        release_artist=release_artist,
                        release_category_str=release_category_str,
                        release_codec=self.get_release_codec_setting(codec_short),
                        release_source_str=release_source_str))

        else:
            """self.category in {ReleaseCategory.ALBUM, ReleaseCategory.ANTHOLOGY, ReleaseCategory.BOOTLEG,
                     ReleaseCategory.CONCERT_RECORDING, ReleaseCategory.DEMO, ReleaseCategory.EP,
                     ReleaseCategory.INTERVIEW, ReleaseCategory.LIVE_ALBUM, ReleaseCategory.REMIX,
                     ReleaseCategory.SINGLE, ReleaseCategory.UNKNOWN}:"""
            return normalize_path_chars(
                "{release_artist} - {year} - {release_name} "
                "{release_category_str}{release_source_str}[{release_codec}]"
                .format(release_artist=release_artist,
                        year=year,
                        release_name=release_name,
                        release_category_str=release_category_str,
                        release_source_str=release_source_str,
                        release_codec=self.get_release_codec_setting(codec_short)))

    def validate_codec(self) -> Optional[str]:
        codec_settings = unique([track.get_codec() for track in self.tracks.values()])
        if len(codec_settings) != 1:
            return None
        return codec_settings[0]

    # return true if the folder name can be validated
    def can_validate_folder_name(self) -> bool:
        return self.validate_release_date() is not None and self.validate_release_title() is not None \
               and self.validate_release_artists() != [] and self.validate_codec()

    # return release date if consistent
    def validate_release_date(self) -> str:
        dates = unique([track.date for track in self.tracks.values()])
        if len(dates) == 1 and dates[0]:
            return dates[0]

    # return number of tracks with empty artist tags
    def blank_artists(self) -> int:
        return sum([1 for track in self.tracks.values() if not len(track.artists)])

    # return number of tracks with empty track title tags
    def blank_track_titles(self) -> int:
        return sum([1 for track in self.tracks.values() if not track.track_title])

    # return release artists if consistent
    def validate_release_artists(self) -> List[str]:
        release_artists = unique([track.release_artists for track in self.tracks.values()])
        if len(release_artists) == 1:
            return unique(release_artists[0])
        return []

    def extract_release_artist(self) -> List[str]:
        artists = unique([x.artists for x in self.tracks.values()])
        return artists[0] if len(artists) == 1 else []

    # return release title if consistent
    def validate_release_title(self) -> Optional[str]:
        release_titles = unique([track.release_title for track in self.tracks.values()])
        if len(release_titles) == 1 and release_titles[0] != "":
            return release_titles[0]

        # if the release title couldn't be validated, try normalizing it
        release_titles = unique([normalize_release_title(track.release_title) for track in self.tracks.values()])
        if len(release_titles) == 1 and release_titles[0] != "":
            return release_titles[0]

    def __get_disc_numbers_by_track(self) -> Dict[int, List[int]]:
        # extract into track_numbers[disc]
        track_numbers = OrderedDict()
        for track in self.tracks.values():
            disc_num = track.disc_number or 1
            if disc_num not in track_numbers:
                track_numbers[disc_num] = []
            if track.track_number:
                track_numbers[disc_num].append(int(track.track_number))

        for disc_num in track_numbers:
            track_numbers[disc_num] = sorted(track_numbers[disc_num])

        return OrderedDict(sorted(track_numbers.items(), key=lambda disc: disc[0]))

    # return empty list if consistent
    def validate_track_numbers(self) -> Dict[int, list]:

        # extract into track_numbers[disc]
        track_numbers = self.__get_disc_numbers_by_track()

        # validate a full set of strictly incrementing tracks, starting at 1
        all_tracks_present = True

        for disc in track_numbers:

            if not track_numbers[disc] or track_numbers[disc][0] != 1:
                all_tracks_present = False

            for i in range(0, len(track_numbers[disc]) - 1):
                if track_numbers[disc][i] != track_numbers[disc][i + 1] - 1:
                    all_tracks_present = False

        if all_tracks_present:
            return OrderedDict()
        else:
            return track_numbers

    def resequence_track_numbers(self):
        """If multiple discs, disc numbers are complete, and track numbers are sequenced as if there is only one disc"""
        track_numbers = self.validate_track_numbers()

        if not track_numbers:
            return

        disc_nums = [x for x in track_numbers]
        if disc_nums[0] != 1:
            return
        for i in range(len(disc_nums) - 1):
            if disc_nums[i] != disc_nums[i+1] - 1:
                return

        flat_track_numbers = []
        for disc in track_numbers:
            flat_track_numbers += track_numbers[disc]

        if not flat_track_numbers or flat_track_numbers[0] != 1:
            return

        for i in range(len(flat_track_numbers) - 1):
            if flat_track_numbers[i] != flat_track_numbers[i+1] - 1:
                return

        curr_track = 1
        curr_disc = 1
        for filename in self.tracks:
            if curr_disc != self.tracks[filename].disc_number:
                curr_track = 1
                curr_disc = self.tracks[filename].disc_number

            self.tracks[filename].track_number = curr_track
            curr_track += 1





    def get_total_tracks(self) -> Dict[int, int]:

        result = OrderedDict()

        track_numbers = self.__get_disc_numbers_by_track()

        for disc in track_numbers:
            all_tracks_present = True
            for i in range(0, len(track_numbers[disc]) - 1):
                if track_numbers[disc][i] != track_numbers[disc][i + 1] - 1:
                    all_tracks_present = False

            result[disc] = len(track_numbers[disc]) if all_tracks_present else None

        return result

    # return an invalid sequence of total tracks tags
    def validate_total_tracks(self) -> List[int]:
        violating_discs = []
        total_tracks = OrderedDict()
        for track in self.tracks.values():
            if track.disc_number not in total_tracks:
                total_tracks[track.disc_number] = []
            total_tracks[track.disc_number].append(track.total_tracks)

        for disc in total_tracks:
            curr_disc = unique(total_tracks[disc])
            if len(curr_disc) != 1 or len(total_tracks[disc]) != curr_disc[0]:
                violating_discs.append(disc)

        return violating_discs

    # return an invalid sequence of disc numbers
    def validate_disc_numbers(self) -> List[int]:
        track_numbers = self.__get_disc_numbers_by_track()
        if not track_numbers:
            return []

        # disc number
        disc_numbers = sorted(list(track_numbers))
        all_discs_present = disc_numbers[0] == 1

        for i in range(0, len(disc_numbers) - 1):
            if disc_numbers[i] != disc_numbers[i + 1] - 1:
                all_discs_present = False

        if all_discs_present:
            return []
        else:
            return disc_numbers

    def validate_total_discs(self) -> bool:
        disc_numbers = sorted(list(self.__get_disc_numbers_by_track()))
        total_discs = unique([track.total_discs for track in self.tracks.values()])
        return len(total_discs) == 1 and len(disc_numbers) and total_discs[0] == disc_numbers[-1]

    def get_total_discs(self) -> Optional[int]:
        disc_numbers = sorted(unique([track.disc_number
                                      for track in self.tracks.values() if track.disc_number is not None]))
        if not len(disc_numbers) or disc_numbers[0] != 1:
            return None
        for x in range(len(disc_numbers) - 1):
            if disc_numbers[x] != disc_numbers[x+1] - 1:
                return None
        return disc_numbers[-1]

    # return a list of genres, or [] if inconsistent/empty
    def validate_genres(self):
        genres = unique([track.genres for track in self.tracks.values()])
        return genres[0] if len(genres) == 1 else []

    # return multiple tag types or empty list
    def get_tag_types(self) -> List[str]:
        return unique([track.stream_info.tag_type for track in self.tracks.values()])

    def get_codecs(self) -> List[str]:
        return unique([track.get_codec_setting_str(short=True) for track in self.tracks.values()])

    def get_cbr_bitrates(self) -> List[int]:
        if (self.get_codecs()[0] if len(self.get_codecs()) == 1 else "") == "CBR":
            return unique([track.stream_info.bitrate for track in self.tracks.values()])
        else:
            return []

    def get_vbr_bitrate(self) -> Optional[float]:
        codecs = self.get_codecs()
        if not len(unique(codecs)) == 1 or codecs[0] != "VBR":
            return None

        return sum([track.stream_info.bitrate * track.stream_info.length for track in self.tracks.values()]) / \
               sum([track.stream_info.length for track in self.tracks.values()])



from distutils.core import setup
setup(
    name = 'metafix',
    packages = ['metafix'],
    version = '1.3.0',
    description = 'Provides validation and safe, automated repair of audio metadata. Supports MP3, FLAC, and other popular formats.',
    url = 'https://github.com/spiritualized/metafix',
    download_url = 'https://github.com/spiritualized/metafix/archive/v1.3.0.tar.gz',
    keywords = ['metadata', 'validation', 'mp3', 'flac', 'python', 'library'],
    install_requires = [
                    'cleartag>=1.2.1',
                    'lastfmcache>=1.2.4',
                    'ordered-set>=3.1.1',
                ],

    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Programming Language :: Python :: 3.6',
    ],
)



from enum import Enum


class ReleaseCategory(Enum):
    ALBUM = 'Album'
    ANTHOLOGY = 'Anthology'
    BOOTLEG = 'Bootleg'
    CDM = 'CDM'
    COMPILATION = 'Compilation'
    CONCERT_RECORDING = 'Concert Recording'
    DEMO = 'Demo'
    EP = 'EP'
    VIDEO_GAME_MUSIC = 'Video Game Music'
    INTERVIEW = 'Interview'
    LIVE_ALBUM = 'Live Album'
    MIX = 'Mix'
    MIXTAPE = 'Mixtape'
    REMIX = 'Remix'
    SINGLE = 'Single'
    SOUNDTRACK = 'Soundtrack'
    UNKNOWN = 'Unknown'


class ReleaseSource(Enum):
    CD = 'CD'
    VINYL = 'Vinyl'
    WEB = 'Web'
    UNKNOWN = 'Unknown'


class ViolationType(Enum):
    ARTIST_WHITESPACE = 'artist-whitespace'
    RELEASE_ARTIST_WHITESPACE = 'release-artist-whitespace'
    DATE_WHITESPACE = 'date-whitespace'
    RELEASE_TITLE_WHITESPACE = 'release-title-whitespace'
    TRACK_TITLE_WHITESPACE = 'track-title-whitespace'
    GENRE_WHITESPACE = 'genre-whitespace'
    DATE_INCONSISTENT = 'date-inconsistent'
    ARTIST_BLANK = 'artist'
    TRACK_TITLE_BLANK = 'track-title'
    RELEASE_ARTIST_INCONSISTENT = 'release-artist'
    RELEASE_ARTIST_SPELLING = 'release-artist-spelling'
    RELEASE_ARTIST_NOT_FOUND = 'release-artist-not-found'
    RELEASE_TITLE_INCONSISTENT = 'release-title-inconsistent'
    RELEASE_TITLE_CATEGORY = 'release-title-category'
    RELEASE_TITLE_SOURCE = 'release-title-source'
    RELEASE_TITLE_SPELLING = 'release-title-spelling'
    DATE_INCORRECT = 'date-incorrect'
    BAD_GENRES = 'bad-genres'
    INCORRECT_TRACK_TITLE = 'incorrect-track-title'
    TRACK_ARTIST_SPELLING = 'track-artist-spelling'
    MISSING_TRACKS = 'missing-tracks'
    TOTAL_TRACKS_INCONSISTENT = 'total-tracks'
    MISSING_DISCS = 'missing-discs'
    TOTAL_DISCS_INCONSISTENT = 'total-discs'
    TAG_TYPES_INCONSISTENT = 'tag-types'
    CODECS_INCONSISTENT = 'codecs'
    CBR_INCONSISTENT = 'cbr'
    FILENAME = 'filename'
    FOLDER_NAME = 'folder-name'
    ARTIST_LOOKUP = 'artist-lookup'
    UNREADABLE = 'unreadable'
    COMMENT_SUBSTRING = 'comment-substring'

upgrade_message = "The application is expired. Please upgrade to the newest version."